import {Logger} from "./logger.js";

const log = Logger.logger("WellKnownData");

// List of attributes for the openID configuration
const FXA_ENDPOINT_PROFILE = "userinfo_endpoint";
const FXA_ENDPOINT_TOKEN = "token_endpoint";
const FXA_ENDPOINT_ISSUER = "issuer";

const FETCH_TIMEOUT = 10000; // 10 secs

// Token expiration time
const FXA_EXP_WELLKNOWN_TIME = 3600; // 1 hour

// A class to fetch well-known data only when needed.
// Well-known data resource is fetched only when required and if the current
// data is expired. The TTL is 1 hour.

export class WellKnownData {
  constructor() {
    this.fxaEndpoints = new Map();
    this.fxaEndpointsReceivedAt = 0;
    this.openIDJSON = {};
  }

  async init() {
    this.fxaOpenID = await ConfigUtils.getFxaOpenID();

    // Let's see if we can fetch data now. But let's ignore the async return
    // value.
    this.fetch();
  }

  syncGetEndpoint(name) {
    return this.fxaEndpoints.get(name);
  }

  syncSetEndpoint(name, value) {
    return this.fxaEndpoints.set(name, value);
  }

  hasWellKnownData() {
    return this.fxaEndpoints.size !== 0;
  }

  async fetch() {
    log("Fetching well-known data");

    let now = Date.now();
    let nowInSecs = Math.floor(now / 1000);

    if ((this.fxaEndpointsReceivedAt + FXA_EXP_WELLKNOWN_TIME) > nowInSecs) {
      log("Well-knonw data cache is good");
      return true;
    }

    log("Fetching well-known data for real");

    // Let's fetch the data with a timeout of FETCH_TIMEOUT milliseconds.
    let json;
    try {
      json = await Promise.race([
        fetch(this.fxaOpenID, {cache: "no-cache"}).then(r => r.json(), e => {
          console.error("Failed to fetch the well-known resource", e);
          return null;
        }),
        new Promise(resolve => {
          setTimeout(_ => resolve(null), FETCH_TIMEOUT);
        }),
      ]);
    } catch (e) {
      console.error("Failed to fetch the well-known resource", e);
    }

    if (!json) {
      return false;
    }

    this.syncSetEndpoint(FXA_ENDPOINT_PROFILE, json[FXA_ENDPOINT_PROFILE]);
    this.syncSetEndpoint(FXA_ENDPOINT_TOKEN, json[FXA_ENDPOINT_TOKEN]);
    this.syncSetEndpoint(FXA_ENDPOINT_ISSUER, json[FXA_ENDPOINT_ISSUER]);

    this.fxaEndpointsReceivedAt = nowInSecs;

    this.openIDJSON = json;
    return true;
  }

  async openID() {
    await this.fetch();
    return this.openIDJSON;
  }

  isAuthUrl(origin) {
    if (new URL(this.fxaOpenID).origin === origin) {
      return true;
    }

    if (!this.hasWellKnownData()) {
      return false;
    }

    // If is part of oauth also ignore
    const authUrls = [
      this.syncGetEndpoint(FXA_ENDPOINT_PROFILE),
      this.syncGetEndpoint(FXA_ENDPOINT_TOKEN),
    ];

    return authUrls.some((item) => {
      return new URL(item).origin === origin;
    });
  }

  excludedDomains() {
    let excludedDomains = [new URL(this.fxaOpenID).hostname];

    if (this.hasWellKnownData()) {
      [FXA_ENDPOINT_PROFILE, FXA_ENDPOINT_TOKEN, FXA_ENDPOINT_ISSUER].forEach(e => {
        try {
          excludedDomains.push(new URL(this.syncGetEndpoint(e)).hostname);
        } catch (e) {}
      });
    }

    return excludedDomains;
  }

  async getIssuerEndpoint() {
    return this.getGenericEndpoint(FXA_ENDPOINT_ISSUER);
  }

  async getProfileEndpoint() {
    return this.getGenericEndpoint(FXA_ENDPOINT_PROFILE);
  }

  async getTokenEndpoint() {
    return this.getGenericEndpoint(FXA_ENDPOINT_TOKEN);
  }

  async getGenericEndpoint(endpoint) {
    await this.fetch();
    return this.syncGetEndpoint(endpoint);
  }
}
